from datetime import datetime

from rsi.session import RSISession
from rsi.conf import DEFAULT_RSI_URL
from rsi.exceptions import RSIException

ROADMAP_ENDPOINT = '/graphql'

DATE_STR_FMT = "%Y-%m-%d"
roadmap_query = [{
    "operationName": "Roadmap",
    "variables": {"startDate": "", "endDate": ""},
    "query": """

query Roadmap($startDate: String!, $endDate: String!, $context: String) {
  roadmap(startDate: $startDate, endDate: $endDate, context: $context) {
    ...Team
    deliverables {
      ...Deliverable
      projects {
        ...Project
        __typename
      }
      timeAllocations {
        ...TimeAllocation
        discipline {
          ...Discipline
          __typename
        }
        __typename
      }
      __typename
    }
    __typename
  }
}

fragment Team on Team {
  title
  description
  __typename
}

fragment Deliverable on Deliverable {
  title
  description
  startDate
  endDate
  __typename
}

fragment Project on Project {
  title
  logo
  __typename
}

fragment TimeAllocation on TimeAllocation {
  startDate
  endDate
  __typename
}

fragment Discipline on Discipline {
  title
  color
  countMembers
  __typename
}
"""
}]


class Roadmap(object):
    def __init__(self, session=None, rsi_url=DEFAULT_RSI_URL, roadmap_endpoint=ROADMAP_ENDPOINT):
        """ Queries information from the RSI Roadmap

        :argument cache_ttl How long to cache the results of the API before re-querying
        """
        self.session = session or RSISession(url=rsi_url)
        self.rsi_url = rsi_url.rstrip('/')
        self.roadmap_endpoint = '{}/{}'.format(self.rsi_url, roadmap_endpoint.lstrip('/'))

        if self.session is None:
            self.session = RSISession(url=rsi_url)

    def fetch_roadmap(self, start_date: datetime, end_date: datetime):
        """

        :param start_date: Datetime beginning of the roadmap to search for
        :param end_date: Datetime end of the roadmap to search for
        :return: diction of roadmap entries
        """
        q = roadmap_query.copy()
        q[0]['variables'].update({
            'startDate': start_date.strftime(DATE_STR_FMT),
            'endDate': end_date.strftime(DATE_STR_FMT)
        })
        try:
            p = self.session.post(self.roadmap_endpoint, json=q)
            if p.status_code == 200:
                p = p.json()[0]
                if p.get('errors', []):
                    raise RSIException(p['errors'])
                return p.get('data', {}).get('roadmap', [])
        except Exception as e:
            raise
        return []



# -*- coding: utf-8 -*-

"""Top-level package for Python RSI."""
__author__ = """Ventorvar"""
__email__ = 'ventorvar@gmail.com'
__version__ = '0.1.16'



# TODO: future endpoints to check out
# https://robertsspaceindustries.com/api/stats/getCrowdfundStats
# https://robertsspaceindustries.com/api/account/badge/getBadges

from .rsi import RSISite



from .session import RSISession

DEFAULT_LAUNCHER_API_ENDPOINTS = {
    'games_claims': '/api/launcher/v3/games/claims',
    'games_library': '/api/launcher/v3/games/library',
    'release': '/api/launcher/v3/games/release',
    'news': '/api/launcher/v3/content/news',
    'patch_notes': '/api/launcher/v3/content/patchNotes',
}


class LauncherAPI:
    def __init__(self, session: RSISession, **kwargs):
        self.session = session

        def _kwargs_or_default(key):
            endpoint = kwargs[key] if key in kwargs else DEFAULT_LAUNCHER_API_ENDPOINTS[key]
            return endpoint.lstrip('/')

        self._games_claims_api = f"{self.session.url}/{_kwargs_or_default('games_claims')}"
        self._games_library_api = f"{self.session.url}/{_kwargs_or_default('games_library')}"
        self._games_release = f"{self.session.url}/{_kwargs_or_default('release')}"
        self._content_news = f"{self.session.url}/{_kwargs_or_default('news')}"
        self._content_patch_notes = f"{self.session.url}/{_kwargs_or_default('patch_notes')}"

        self._games_claims = None
        self._games_library = None

    @property
    def claims(self):
        if self.session.is_authenticated:
            if self._games_claims is None:
                success, info = self.session.query_api(self._games_claims_api)
                if success:
                    self._games_claims = info['data']
            return self._games_claims
        return None

    @property
    def library(self):
        if self.claims is not None:
            if self._games_library is None:
                success, info = self.session.query_api(self._games_library_api, json={"claims": self.claims})
                if success:
                    self._games_library = info['data']
            return self._games_library
        return None

    def news(self, game_id="SC"):
        success, info = self.session.query_api(self._content_news, json={"game_id": game_id})
        if success:
            return info['data']
        raise ValueError(f'{info}')

    def patch_notes(self, game_id="SC", channel_id="LIVE"):
        success, info = self.session.query_api(self._content_patch_notes,
                                               json={"game_id": game_id, "channel_id": channel_id})
        if success:
            return info['data']
        raise ValueError(f'{info}')

    def release(self, game_id="SC", channel_id="LIVE"):
        success, info = self.session.query_api(self._games_release,
                                               json={"claims": self.claims, "gameId": game_id, "channelId": channel_id})
        if success:
            return info['data']
        raise ValueError(f'{info}')



import re
from cachetools import TTLCache

from bs4 import BeautifulSoup
from rsi.session import RSISession
from rsi.conf import DEFAULT_RSI_URL
from rsi.exceptions import RSIException

PLEDGE_SKU_ENDPOINT = '/api/store/getSKUs'
SHIP_UPGRADE_ENDPOINT = '/pledge-store/api/upgrade'
SET_CONTEXT_TOKEN_ENDPOINT = '/pledge-store/api/setContextToken'
SHIP_UPGRADE_RE = re.compile(r'fromShips: (\[.*\]), toShips')


upgrades_initShipUpgrades_query = [{
    "operationName": "initShipUpgrade", "variables": {},
    "query": """
query initShipUpgrade {
  ships {
    id
    name
    medias {
      productThumbMediumAndSmall
      slideShow
    }
    manufacturer {
      id
      name
    }
    focus
    type
    flyableStatus
    owned
    msrp
    link
    skus {
      id
      title
      available
      price
      body
    }
  }
  manufacturers {
    id
    name
  }
  app {
    version
    env
    cookieName
    sentryDSN
    pricing {
      currencyCode
      currencySymbol
      exchangeRate
      taxRate
      isTaxInclusive
    }
    mode
    isAnonymous
    buyback {
      credit
    }
  }
}
"""}]


SKU_PRODUCT_IDS = {
    'standalone_ships': 72,
    'paints': 268,
    'gift_cards': 60,
    'addons': 3,
    'uec': 41,
    'event_tickets': 67,
    'digital_goodies': 222

}


class PledgeStore(object):
    def __init__(self, session=None, rsi_url=DEFAULT_RSI_URL, sku_endpoint=PLEDGE_SKU_ENDPOINT, cache_ttl=300,
                 ship_upgrade_endpoint=SHIP_UPGRADE_ENDPOINT, set_context_token_endpoint=SET_CONTEXT_TOKEN_ENDPOINT):
        """ Queries information from the RSI pledge store.

        :argument cache_ttl How long to cache the results of the API before re-querying
        """
        self.session = session or RSISession(url=rsi_url)
        self.rsi_url = rsi_url.rstrip('/')
        self.sku_endpoint = '{}/{}'.format(self.rsi_url, sku_endpoint.lstrip('/'))
        self.ship_upgrade_endpoint = '{}/{}'.format(self.rsi_url, ship_upgrade_endpoint.lstrip('/'))
        self._set_context_token_endpoint = '{}/{}'.format(self.rsi_url, set_context_token_endpoint.lstrip('/'))
        self._ttlcache = TTLCache(maxsize=1, ttl=cache_ttl)

        if self.session is None:
            self.session = RSISession(url=rsi_url)

    def skus(self, product_id="", search="", storefront="pledge", type="", sort='price_desc', pages=9999):
        page = 1
        opts = {"product_id": product_id, "search": search, "storefront": storefront, "type": type, "sort": sort}
        r = self.session.post(self.sku_endpoint, json={**opts, **dict(page=page)})
        r.raise_for_status()
        r = r.json()
        if not r['success']: return {}

        row_count = r['data']['rowcount']
        html = r['data']['html']
        while row_count < r['data']['totalrows'] and page < pages:
            page += 1
            r = self.session.post(self.sku_endpoint, json={**opts, **dict(page=page)})
            r.raise_for_status()
            r = r.json()
            if not r['success']: return {}
            html += r['data']['html']
            row_count += r['data']['rowcount']

        soup = BeautifulSoup(html, features='html.parser')
        for item in soup.select('div.product-item.js-ecommerce-tracking-sku'):
            try:
                yield item.select('.title')[0].text.strip(), dict(
                    title=item.select('.title')[0].text.strip(),
                    image=item.select('img')[0].get('src', ''),
                    price=item.select('.final-price')[0].get('data-value', ''),
                    price_str=item.select('.final-price')[0].text.strip(),
                    stock=item.select('.state')[0].text.strip(),
                    link=f'{self.rsi_url}{item.select(".more")[0].get("href", "")}',
                )
            except Exception as e:
                print(repr(e))

    def pledge_extras(self, product_id="", search="", *args, **kwargs):
        return self.skus(product_id, search, type="extras", *args, **kwargs)

    def pledge_game_packages(self, product_id="", search="", *args, **kwargs):
        return self.skus(product_id, search, type="game-packages")

    def ship_upgrades(self):
        pledge_map = {}
        try:
            self.session.update_session_tokens(extra=[self._set_context_token_endpoint])
            p = self.session.post(self.ship_upgrade_endpoint, json=upgrades_initShipUpgrades_query)
            if p.status_code == 200:
                p = p.json()[0]
                if p.get('errors', []):
                    raise RSIException(p['errors'])
                return {_['id']: _ for _ in p.get('data', {}).get('ships', [])}
        except Exception as e:
            raise
        return pledge_map




def get_item(iterable_or_dict, index, default=None):
    """Return iterable[index] or default if IndexError is raised."""
    try:
        return iterable_or_dict[index]
    except (IndexError, KeyError):
        return default



from .session import RSISession
from .pledge_store import PledgeStore
from .shipmatrix import ShipMatrixAPI
from .org import OrgAPI
from .citizen import fetch_citizen
from .status import Status
from .roadmap import Roadmap


class RSISite:
    def __init__(self, session: RSISession = None, *args, **kwargs):
        self.session = session
        if self.session is None:
            self.session = RSISession(*args, **kwargs)

        self.store = PledgeStore(session=self.session)
        self.ships = ShipMatrixAPI(session=self.session)
        self.roadmap = Roadmap(session=self.session)
        self.status = Status()

    @property
    def is_authenticated(self):
        return self.session.is_authenticated

    def authenticate(self, username, password, force=False):
        return self.session.authenticate(username, password, force=force)

    def citizen(self, handle, skip_orgs=False):
        return fetch_citizen(handle, skip_orgs=skip_orgs, session=self.session)

    def org(self, symbol):
        return OrgAPI(symbol=symbol, session=self.session)



import os
import codecs
import pickle
import requests
import configparser

from rsi.conf import DEFAULT_RSI_URL

DEFAULT_SESSION_CONFIG = {
    'name': '',
    'id': '',
    'cookies': '',
}

RSI_SESSION_DURATION = ['session', 'day', 'week', 'month', 'year']

DEFAULT_API_ENDPOINTS = {
    'set_auth_token_endpoint': '/api/account/v2/setAuthToken',
    'login_endpoint': '/api/launcher/v3/signin',
    'login_two_factor_endpoint': '/api/account/signinMultiStep',
    'session_check_endpoint': '/api/launcher/v3/games/claims',
    'signout_endpoint': '/api/account/signout',
}


def cli_two_factor_prompt():
    code = input('RSI Authenticator Code: ')
    try:
        if len(code) != 6:
            raise ValueError()
        int(code)
    except ValueError:
        raise ValueError('Invalid code entered: {}'.format(code))
    return code


class RSISession(requests.Session):
    def __init__(self, url=DEFAULT_RSI_URL, persist_session=True, session_file='.pyrsi_session', clear_session=False,
                 allow_two_factor=True, two_factor_prompt=cli_two_factor_prompt, two_factor_duration='session',
                 username=None, password=None, **kwargs):
        super(RSISession, self).__init__()

        self.hooks['response'].append(self._update_rsi_token)
        self.url = url.rstrip('/')

        def _kwargs_or_default(key):
            endpoint = kwargs[key] if key in kwargs else DEFAULT_API_ENDPOINTS[key]
            return endpoint.lstrip('/')

        self._login_api = f"{self.url}/{_kwargs_or_default('login_endpoint')}"
        self._login_two_factor_api = f"{self.url}/{_kwargs_or_default('login_two_factor_endpoint')}"
        self._session_check_api = f"{self.url}/{_kwargs_or_default('session_check_endpoint')}"
        self._signout_api = f"{self.url}/{_kwargs_or_default('signout_endpoint')}"
        self._set_auth_token = f"{self.url}/{_kwargs_or_default('set_auth_token_endpoint')}"

        self._allow_two_factor = allow_two_factor
        self.two_factor_prompt = two_factor_prompt
        self.two_factor_duration = two_factor_duration

        self._config = configparser.ConfigParser()
        self._config.add_section('RSI')
        self.session_file = session_file
        self.persist_session = persist_session
        self.session_name = 'RSI-Token'
        self.session_id = ''

        if clear_session:
            self.clear_session()

        if self.persist_session:
            self._load_session()

        if username is not None and password is not None:
            self.authenticate(username, password)

    def _load_session(self):
        self._config.read(self.session_file)
        cookies = self._config.get('RSI', 'cookies', fallback='')
        if cookies:
            self.cookies.update(pickle.loads(codecs.decode(cookies.encode(), "base64")))
        self._update_session(self._config.get('RSI', 'session_name', fallback=''),
                             self._config.get('RSI', 'session_id', fallback=''),
                             save=False)

    def _update_session(self, name, id, save=True):
        if not id:
            # if id has been cleared, clear the session
            if name in self.headers:
                del self.headers[name]
            self.cookies.clear()
            self.session_name = self.session_id = ''
        else:
            self.session_name = name
            self.session_id = id
            self.headers.update({'X-{}'.format(name): id})

        self._config.set('RSI', 'session_name', name)
        self._config.set('RSI', 'session_id', id)
        self._config.set('RSI', 'cookies', codecs.encode(pickle.dumps(self.cookies), "base64").decode())
        if save and self.persist_session:
            with open(self.session_file, 'w') as f:
                self._config.write(f)

    def _update_rsi_token(self, response, *args, **kwargs):
        # called on the response hook and will update the session id from the cookie
        if self.session_name in response.cookies:
            if response.cookies[self.session_name] != self.session_id:
                self._update_session(self.session_name, self.session_id)

    def query_api(self, api, json=None):
        resp = self.post(api, json=json)
        info = {}

        if resp.status_code == 200:
            info = resp.json()
            if info.get('success', 0) == 1:
                return True, info
        return False, info

    def clear_session(self):
        if os.path.isfile(self.session_file):
            os.remove(self.session_file)

    def update_session_tokens(self, extra=[]):
        """ extra is a list of extra set*Token APIs you'd like to be hit """
        self.query_api(self._set_auth_token)
        for _ in extra:
            self.query_api(_)

    @property
    def is_authenticated(self):
        is_auth, _ = self.query_api(self._session_check_api)
        return is_auth

    def signout(self):
        success, _ = self.query_api(self._signout_api)
        self.clear_session()
        return success

    def authenticate(self, username, password, force=False):
        if not force and self.is_authenticated:
            return

        self._update_session(self.session_name, '')

        success, info = self.query_api(self._login_api, json={
            'username': username,
            'password': password,
            'remember': 'off'})

        if not success and info.get('code', '') == 'ErrMultiStepRequired':
            # Two factor auth required, ask user for
            if not self._allow_two_factor:
                raise Exception('Account requires Two Factor authentication which has been disabled')
            self._update_session(info['data']['session_name'], info['data']['session_id'])
            code = self.two_factor_prompt()
            success, info = self.query_api(self._login_two_factor_api, json={
                'code': code,
                'device_name': 'pyrsi',
                'device_type': 'computer',
                'duration': self.two_factor_duration
            })

        if not success:
            raise Exception('Unable to log in: {}'.format(info))
        self._update_session(info['data']['session_name'], info['data']['session_id'])

        return True



import re as _re

from bs4 import BeautifulSoup as _bs
from rsi.utils import get_item
from rsi.conf import DEFAULT_RSI_URL
from rsi.session import RSISession


def fetch_citizen(name, url=DEFAULT_RSI_URL, endpoint='/citizens', skip_orgs=False, session=None):
    session = session or RSISession()
    result = {}
    url = url.rstrip('/')
    citizen_url = "{}/{}/{}".format(url.rstrip('/'), endpoint.strip('/'), name)
    orgapiurl = '{}/{}'.format(url.rstrip('/'), 'api/orgs/getOrgMembers')

    page = session.get(citizen_url)
    if page.status_code == 200:
        soup = _bs(page.text, features='html.parser')
        _ = [_.text for _ in soup.select(".info .value")[:3]]
        result['username'] = get_item(_, 0, '')
        result['handle'] = get_item(_, 1, '')
        result['title'] = get_item(_, 2, '')
        result['title_icon'] = get_item(soup.select(".info .icon img"), 0, '')
        if result['title_icon']:
            result['title_icon'] = '{}/{}'.format(url, result['title_icon']['src'])
        result['avatar'] = "{}/{}".format(url, soup.select('.profile .thumb img')[0]['src'].lstrip('/'))
        result['url'] = citizen_url

        if soup.select('.profile-content .bio'):
            result['bio'] = soup.select('.profile-content .bio')[0].text.strip('\nBio').strip()
        else:
            result['bio'] = ''
        result['citizen_record'] = soup.select('.citizen-record .value')[0].text
        try:
            result['citizen_record'] = int(result['citizen_record'][1:])
        except:
            pass

        _ = {_.select_one('span').text:
             _re.sub(r'\s+', ' ', _.select_one('.value').text.strip()).replace(' ,', ',')
             for _ in soup.select('.profile-content > .left-col .entry')}
        result['enlisted'] = get_item(_, 'Enlisted', '')
        result['location'] = get_item(_, 'Location', '')
        result['languages'] = get_item(_, 'Fluency', '')
        result['languages'] = result['languages'].replace(',', '').split()

        if not skip_orgs:
            orgs_page = session.get("{}/organizations".format(citizen_url))
            if orgs_page.status_code == 200:
                orgsoup = _bs(orgs_page.text, features='html.parser')
                result['orgs'] = []
                for org in orgsoup.select('.orgs-content .org'):
                    orgname, sid, rank = [_.text for _ in org.select('.info .entry .value')]
                    if orgname[0] == '\xa0':
                        orgname = sid = rank = 'REDACTED'

                    roles = []
                    r = session.post(orgapiurl, data={'symbol': sid, 'search': name})
                    if r.status_code == 200:
                        r = r.json()
                        if r['success'] == 1:
                            apisoup = _bs(r['data']['html'], features='html.parser')
                            roles = [_.text for _ in apisoup.select('.rolelist .role')]

                    orgdata = {
                        'name': orgname,
                        'sid': sid,
                        'rank': rank,
                        'roles': roles,
                    }
                    try:
                        orgdata['icon'] = '{}/{}'.format(url, org.select('.thumb img')[0]['src'].lstrip('/'))
                    except IndexError:
                        pass

                    result['orgs'].append(orgdata)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "cachetools",
    "requests",
    "fuzzywuzzy",
    "python-Levenshtein",
    "beautifulsoup4",
    "configparser"
]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Ventorvar",
    author_email='ventorvar@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
    description="Python API for interacting with the Roberts Space Industries site for Star Citizen.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='pyrsi',
    name='pyrsi',
    packages=find_packages(include=['rsi']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitlab.com/scmodding/frameworks/pyrsi',
    version='0.1.16',
    zip_safe=True,
)



import re
import time
from collections import defaultdict
from fuzzywuzzy import process
from cachetools import TTLCache
from bs4 import BeautifulSoup
from rsi.conf import DEFAULT_RSI_URL
from rsi.session import RSISession
from rsi.pledge_store import PledgeStore
from rsi.exceptions import RSIException

DEFAULT_SHIPMATRIX_ENDPOINT = '/ship-matrix/index'
DEFAULT_LOANER_MATRIX_URL = 'https://support.robertsspaceindustries.com/hc/en-us/articles/360003093114-Loaner-Ship-Matrix'
SHIP_MODEL_RE = re.compile(r"model_3d:\s*'(\S+)'")
DEFAULT_CACHE_TTL = 300
SHIP_UPGRADE_URL = 'https://robertsspaceindustries.com/pledge/ship-upgrades'
SHIP_UPGRADE_RE = re.compile(r'fromShips: (\[.*\]), toShips')


class ShipMatrixAPI(object):
    def __init__(self, session=None, rsi_url=DEFAULT_RSI_URL, api_endpoint=DEFAULT_SHIPMATRIX_ENDPOINT, cache_ttl=300,
                 enable_pledges=True, enable_ship_models=True,
                 loaner_ship_url=DEFAULT_LOANER_MATRIX_URL):
        """ Queries information from the RSI Ship Matrix.

        :argument api_endpoint The URL to use to connect to the ship matrix API
        :argument cache_ttl How long to cache the results of the API before re-querying
        """
        self.session = session or RSISession()
        self.rsi_url = rsi_url.rstrip('/')
        self.api_endpoint = '{}/{}'.format(self.rsi_url, api_endpoint.lstrip('/'))
        self._enable_pledges = enable_pledges
        self._enable_ship_models = enable_ship_models
        self._loaner_ship_url = loaner_ship_url
        self._ttlcache = TTLCache(maxsize=3, ttl=cache_ttl)

    def clear_cache(self):
        """ Resets the cache """
        del self._ttlcache['ships_by_name']
        del self._ttlcache['ships']
        del self._ttlcache['loaners']

    def _fuzzy_choices(self):
        return {k: v['name'] for k, v in self.ships.items()}

    def _update_loaner_cache(self):
        p = self.session.get(self._loaner_ship_url)
        p.raise_for_status()

        def _lookup_by_name(name):
            return process.extractBests(name, self._fuzzy_choices(), score_cutoff=80)

        loaners = defaultdict(set)
        soup = BeautifulSoup(p.text, features='html.parser')
        for row in soup.select('.article-body table tbody tr'):
            your_ship, our_loaners = [_.text for _ in row.select('td')]
            our_loaners = [_.strip() for _ in our_loaners.split(',')]

            if your_ship in self.ships_by_name:
                ships = [your_ship]
            elif ' Series' in your_ship:
                ships = [_[0] for _ in _lookup_by_name(your_ship.replace(' Series', ''))]
            elif ' / ' in your_ship:
                ships = []
                for ship in your_ship.split(' / '):
                    ships.append(

                    )
                    ships = [_[0] for _ in _lookup_by_name(ships) if _]

            if not ships:
                print(f'WARNING [update_loaners] could not find ship for {your_ship}')
                continue

            for loaner in our_loaners:
                if loaner.lower() == 'cyclone (explorer only)':
                    for ship in ships:
                        if 'Explorer' in ship:      # handle the 6oo series
                            loaners[ship].update(['Cyclone-RN'])
                else:
                    for ship in ships:
                        loaners[ship].update([_[0] for _ in _lookup_by_name(loaner)])
        self._ttlcache['loaners'] = {k: list(v) for k, v in loaners.items()}

    def _update_ship_cache(self):
        resp = self.session.get(self.api_endpoint)
        resp.raise_for_status()
        data = resp.json()
        if data['msg'] != 'OK':
            raise RSIException(repr(data))
        data = data['data']
        data = {int(_['id']): _ for _ in data}

        pledge_map = {}
        if self._enable_pledges:
            pledges = PledgeStore(session=self.session)
            pledge_map = pledges.ship_upgrades()

        for ship_id in data.keys():
            # rename cargo capacity to be in line with other vars
            if 'cargocapacity' in data[ship_id]:
                data[ship_id]['cargo_capacity'] = data[ship_id].pop('cargocapacity')

            data[ship_id]['pledge_cost'] = pledge_map.get(ship_id, {}).get('msrp', '')
            data[ship_id]['url'] = f'{self.rsi_url}{data[ship_id]["url"]}'
            if data[ship_id]['media']:
                data[ship_id]['media'][0]['source_url'] = f'{self.rsi_url}{data[ship_id]["media"][0]["source_url"]}'
                for _ in data[ship_id]['media'][0]['images']:
                    data[ship_id]['media'][0]['images'][_] = f'{self.rsi_url}{data[ship_id]["media"][0]["images"][_]}'

            if self._enable_ship_models:
                try:
                    p = self.session.get(data[ship_id]['url'])
                    if p.status_code == 200:
                        m = SHIP_MODEL_RE.search(p.text)
                        data[ship_id]['model_3d'] = m.group(1) if m else ''
                except Exception as e:
                    print(f'WARNING: could not lookup ship model for {ship_id} ({data[ship_id]["name"]})')
        self._ttlcache['ships'] = data
        self._ttlcache['ships_by_name'] = {v['name']: v for k, v in data.items()}

    def _from_cache(self, item):
        if item == 'loaners' and 'loaners' not in self._ttlcache:
            self._update_loaner_cache()
        elif 'ships' not in self._ttlcache:
            self._update_ship_cache()
        return self._ttlcache[item]

    @property
    def loaners(self):
        return self._from_cache('loaners')

    @property
    def ships_by_name(self):
        return self._from_cache('ships_by_name')

    @property
    def ships(self):
        return self._from_cache('ships')

    def by_id(self, id):
        return self.ships[id]

    @property
    def names(self):
        return [_['name'] for _ in self.ships]

    def search_by_name(self, ship_name, score_cutoff=80, limit=None):
        """
        Return ships that match the given ship_name using fuzzy matching.

        :param ship_name: Name to match
        :param score_cutoff: minimum matching score to return
        :param limit: limit the number of matches found
        :return: List of matched results in the form of [(dict, int)] where dict is the ship data and in is the
                 matching confidence
        """
        return [(self.ships[_[2]], _[1]) for _ in process.extractBests(ship_name, self._fuzzy_choices(),
                                                                       score_cutoff=score_cutoff, limit=limit)]


if __name__ == "__main__":
    import json
    s = ShipMatrixAPI(enable_ship_models=False)
    print(len(s.ships))
    # print(json.dumps(s.ships, indent=4))
    print(json.dumps(s.loaners, indent=4))



import time
from fuzzywuzzy import process
from cachetools import TTLCache
from bs4 import BeautifulSoup

from rsi.conf import DEFAULT_RSI_URL
from .session import RSISession


DEFAULT_CACHE_TTL = 300


class OrgAPI(object):
    def __init__(self, symbol, session=None, admin_mode=False, url=DEFAULT_RSI_URL, endpoint='/orgs',
                 members_endpoint='/api/orgs/getOrgMembers', cache_ttl=DEFAULT_CACHE_TTL):
        self.symbol = symbol
        self.url = url.rstrip('/')
        self.endpoint = endpoint
        self.members_endpoint = members_endpoint
        self.admin_mode = admin_mode
        self.session = session or RSISession(url=url)

        self.org_url = "{}/{}/{}".format(self.url, self.endpoint.lstrip('/'), symbol)
        self.members_api = "{}/{}".format(self.url, self.members_endpoint.lstrip('/'))
        self._ttlcache = TTLCache(maxsize=1, ttl=cache_ttl)

        self._update_details()   # pull and cache the org details which will raise 404 if not found

    def clear_cache(self):
        """ Resets the cache """
        for key in self._ttlcache.keys():
            del self._ttlcache[key]

    def _cache(self, key, update_func, *args, **kwargs):
        if key not in self._ttlcache:
            self._ttlcache[key] = update_func(*args, **kwargs)
        return self._ttlcache[key]

    def _update_members(self, search):
        members = []

        params = {
            'symbol': self.symbol,
            'search': search,
            'page': 1
        }

        if self.admin_mode:
            params['admin_mode'] = 1

        # this just gets us going
        totalsize = 1
        members_scanned = 0

        while members_scanned < totalsize:
            r = self.session.post(self.members_api, data=params)

            if r.status_code == 200:
                r = r.json()
                if r is None:
                    continue

                if 'data' in r and r['data'] and 'totalrows' in r['data']:
                    totalsize = int(r['data']['totalrows'])

                if r['success'] == 1:
                    apisoup = BeautifulSoup(r['data']['html'], features='html.parser')
                    for member in apisoup.select('.member-item'):
                        members_scanned += 1
                        if member.select('.member-visibility-restriction'):
                            print('skipping hidden member')
                            continue

                        members.append({
                            'name': member.select_one('.name').text,
                            'handle': member.select_one('.nick').text,
                            'avatar': '{}{}'.format(self.url, member.select_one('img').attrs['src']),
                            'affiliate': member.select_one('.title').text == 'Affiliate',
                            'rank': member.select_one('.rank').text,
                            'roles': [_.text for _ in member.select('.rolelist .role')],
                            'url': '{}{}'.format(self.url, member.select_one('a.membercard').attrs['href']),

                            # defaults for things online admins will be able to get the real values of
                            'id': '',
                            'visibility': 'Membership: Visible',
                            'last_online': '',
                        })

                        if self.admin_mode:
                            members[-1].update({
                                'id': member.attrs.get('data-member-id', ''),
                                'last_online': member.select_one('.frontinfo .lastonline').text,
                                'visibility': member.select_one('.frontinfo .visibility').text,
                            })

                    params['page'] = params['page'] + 1
                else:
                    raise ValueError('Received error fetching Org members: {}'.format(r))
            else:
                raise Exception('Received error fetching Org members: {}'.format(r.status_code))
            time.sleep(0.5)
        return members

    def _update_details(self):
        r = self.session.get(self.org_url)
        data = {}
        r.raise_for_status()

        orgsoup = BeautifulSoup(r.text, features='html.parser')
        data['banner'] = '{}{}'.format(self.url, orgsoup.select_one('.banner img')['src'])
        data['logo'] = '{}{}'.format(self.url, orgsoup.select_one('.logo img')['src'])
        data['name'], data['symbol'] = orgsoup.select_one('.inner h1').text.split(' / ')
        data['model'] = orgsoup.select_one('.inner .tags .model').text
        data['commitment'] = orgsoup.select_one('.inner .tags .commitment').text
        data['primary_focus'] = orgsoup.select_one('.inner .focus .primary img')['alt']
        data['secondary_focus'] = orgsoup.select_one('.inner .focus .secondary img')['alt']
        data['join_us'] = orgsoup.select_one('.join-us .body').text.strip()
        return data

    def search(self, handle, score_cutoff=80, limit=None):
        """
        Return members that match the given handle using fuzzy matching.

        :param handle: Handle to match
        :param score_cutoff: minimum matching score to return
        :param limit: limit the number of matches found
        :return: List of matched results in the form of [(dict, int)] where dict is the ship data and in is the
                 matching confidence
        """
        choices = {i: _['handle'] for i, _ in enumerate(self.members)}
        return [(self.members[_[2]], _[1]) for _ in process.extractBests(handle, choices,
                                                                         score_cutoff=score_cutoff, limit=limit)]

    def search_one(self, handle):
        """
        Return the first member that matches the given handle using fuzzy matching, or None

        :param handle: Handle to match
        :return: The best matching member, or None
        """
        choices = self.search(handle, limit=1)
        if choices:
            return choices[0][0]
        return None

    @property
    def members(self):
        return self._cache('members', self._update_members, search='')

    @property
    def details(self):
        return self._cache('details', self._update_details)

    @property
    def banner(self):
        return self.details['banner']

    @property
    def logo(self):
        return self.details['logo']

    @property
    def name(self):
        return self.details['name']

    @property
    def model(self):
        return self.details['model']

    @property
    def commitment(self):
        return self.details['commitment']

    @property
    def primary_focus(self):
        return self.details['primary_focus']

    @property
    def secondary_focus(self):
        return self.details['secondary_focus']

    @property
    def spectrum_url(self):
        return '{}/spectrum/community/{}'.format(self.url, self.symbol)

    @property
    def join_us(self):
        return self.details['join_us']



import requests

DEFAULT_STATUS_API_URL = 'https://status.robertsspaceindustries.com/static/content/api/v0'


class Status:
    """
    Interface to the RSI status page: https://status.robertsspaceindustries.com
    """

    def __init__(self, status_api_url=DEFAULT_STATUS_API_URL, language='en'):
        self.api_url = status_api_url.rstrip('/')
        self.language = language

    def _get(self, endpoint, language, *args, **kwargs):
        lang_map = {'language': language if language is not None else self.language}
        req_url = f'{self.api_url}/{endpoint.format_map(lang_map).lstrip("/")}'
        r = requests.get(req_url, *args, **kwargs)
        r.raise_for_status()
        try:
            return r.json()
        except ValueError:
            return {'error': {'message': 'Could not decode JSON object. Language not available or invalid URL',
                              'url': req_url}}

    def system(self, language=None):
        """
            The current system status.

            :param language:  If specified, override the set language of the `Status` for this query.
        """
        return self._get('/systems.{language}.json', language)

    def timeline(self, language=None):
        """
            The timeline of incidents for the past 7 days.

            :param language:  If specified, override the set language of the `Status` for this query.
        """
        return self._get('/incidents/timeline.{language}.json', language)

    def incident(self, incident_id, language=None):
        """
            Fetch information for a specific incident

            :param incident_id:  Unique ID for a given incident
            :param language:  If specified, override the set language of the `Status` for this query.
        """
        return self._get(f'/incidents/{incident_id}.{{language}}.json', language)
